
/*
 *   OpenGL Extension Module for pdp - pbuffer packet implementation
 *   Copyright (c) by Tom Schouten <pdp@zzz.kotnet.org>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

/*
  this code uses glx. i don't know if it is worth to take into
  account portabiliy. since it will take a while until pdp runs
  on anything else than linux. but in any case, providing a windows/osx
  implementation here should not be too difficult..
*/

#include "pdp_3Dcontext.h"
#include <GL/gl.h>
//#include <GL/glx.h>
#include <GL/glu.h>
//#include <GL/glut.h>


/* constructor */

/* pbuf operators */

u32 pdp_packet_3Dcontext_width(int packet)
{
    t_3Dcontext *c = pdp_packet_3Dcontext_info(packet);
    if (c) return c->width;
    else return 0;
}

u32 pdp_packet_3Dcontext_subwidth(int packet)
{
    t_3Dcontext *c = pdp_packet_3Dcontext_info(packet);
    if (c) return c->sub_width;
    else return 0;
}

u32 pdp_packet_3Dcontext_height(int packet)
{
    t_3Dcontext *c = pdp_packet_3Dcontext_info(packet);
    if (c) return c->height;
    else return 0;
}

u32 pdp_packet_3Dcontext_subheight(int packet)
{
    t_3Dcontext *c = pdp_packet_3Dcontext_info(packet);
    if (c) return c->sub_height;
    else return 0;
}

float pdp_packet_3Dcontext_subaspect(int packet)
{
    t_3Dcontext *c = pdp_packet_3Dcontext_info(packet);
    if (c) return (float)c->sub_width/c->sub_height;
    else return 0;
}

int pdp_packet_3Dcontext_isvalid(int packet)
{
    t_pdp *header = pdp_packet_header(packet);
    t_3Dcontext *c = pdp_packet_3Dcontext_info(packet);

    if (!header) return 0;
    if (!c) return 0;
    if (PDP_3DCONTEXT != header->type) return 0;
    return 1;
}

t_3Dcontext *pdp_packet_3Dcontext_info(int packet)
{
    t_pdp *header = pdp_packet_header(packet);
    if (!header) return 0;
    if (PDP_3DCONTEXT != header->type) return 0;
    return (t_3Dcontext *)&header->info.raw;
}






/* move these to the pdp_3d_context object: they're too specific */

/* setup for 2d operation from pbuf dimensions */
void pdp_packet_3Dcontext_setup_2d_context(int p)
{
    u32 w;
    u32 h;
    float asp;
    if (!pdp_packet_3Dcontext_isvalid(p)) return;
    w = pdp_packet_3Dcontext_subwidth(p);
    h = pdp_packet_3Dcontext_subheight(p);
    asp = pdp_packet_3Dcontext_subaspect(p);


    /* set the viewport to the size of the sub frame */
    glViewport(0, 0, w, h);

    /* set orthogonal projection, with a relative frame size of (2asp x 2) */
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluOrtho2D(0, 2*asp, 0, 2);

    /* set the center of view */
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    glTranslatef(asp, 1, 0);
    glScalef(1,-1,1);

}

/* setup for 3d operation from pbuf dimensions */
void pdp_packet_3Dcontext_setup_3d_context(int p)
{
    u32 w;
    u32 h;
    float asp;
    float m_perspect[] = {-1.f, /* left */
			  1.f,  /* right */
			  -1.f, /* bottom */
			  1.f,  /* top */
			  1.f,  /* front */
			  20.f};/* back */

    if (!pdp_packet_3Dcontext_isvalid(p)) return;
    w = pdp_packet_3Dcontext_subwidth(p);
    h = pdp_packet_3Dcontext_subheight(p);
    asp = pdp_packet_3Dcontext_subaspect(p);


    /* set the viewport to the size of the sub frame */
    glViewport(0, 0, w, h);

    /* set orthogonal projection, with a relative frame size of (2asp x 2) */
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glFrustum(m_perspect[0] * asp, m_perspect[1] * asp,       // left, right
	      m_perspect[2], m_perspect[3],                   // bottom, top
	      m_perspect[4], m_perspect[5]);                  // front, back
    
    /* set the center of view */
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    gluLookAt(0, 0, 4, 0, 0, 0, 0, 1, 0);
    //glTranslatef(asp, 1, 0);


    /* compensate for texture layout in pdp_glx2: $$$TODO fix this!!! */

}


void pdp_3Dcontext_common_setup(void)
{
}
